@@ -3,7 +3,7 @@
 //! This is a private module. Its public items are re-exported by the parent.
 
 use std::{fmt, io};
-use crate::decode;
+use crate::decode::{DecodeError, Source};
 
 
 //------------ Tag -----------------------------------------------------------
@@ -22,8 +22,8 @@ use crate::decode;
 ///
 /// # Limitations
 ///
-/// We can only decode up to four identifier octets. That is, we only support tag
-/// numbers between 0 and 1fffff.
+/// We can only decode up to four identifier octets. That is, we only support
+/// tag numbers between 0 and 1fffff.
 ///
 /// [`Primitive`]: decode/struct.Primitive.html
 /// [`Constructed`]: decode/struct.Constructed.html
@@ -57,7 +57,7 @@ impl Tag {
     ///
     /// (1 bit – 0b1000_0000, it is cleared in the last octet).
     const LAST_OCTET_MASK: u8 = 0x80;
-    
+
     /// The largest tag number possible with three octets.
     const MAX_VAL_SPAN_3_OCTETS: u32 = 0x001f_ffff;
 
@@ -118,18 +118,21 @@ impl Tag {
     /// The tag for the REAL type, UNIVERSAL 9.
     pub const REAL: Self = Tag([9, 0, 0, 0]);
 
-    /// The tag for the ENUMERATED type, UNIVERAL 10.
+    /// The tag for the ENUMERATED type, UNIVERSAL 10.
     pub const ENUMERATED: Self = Tag([10, 0, 0, 0]);
 
-    /// The tag for the EMBEDDED PDV type, UNIVERAL 11.
+    /// The tag for the EMBEDDED PDV type, UNIVERSAL 11.
     pub const EMBEDDED_PDV: Self = Tag([11, 0, 0, 0]);
 
     /// The tag for the UTF8String type, UNIVERSAL 12
     pub const UTF8_STRING: Self = Tag([12, 0, 0, 0]);
 
-    /// The tag for the RELATIVE-OID type, UNIVERAL 13.
+    /// The tag for the RELATIVE-OID type, UNIVERSAL 13.
     pub const RELATIVE_OID: Self = Tag([13, 0, 0, 0]);
 
+    /// The tag for the TIME type, UNIVERSAL 14.
+    pub const TIME: Self = Tag([14, 0, 0, 0]);
+
     /// The tag for the SEQUENCE and SEQUENCE OF types, UNIVERSAL 16.
     pub const SEQUENCE: Self = Tag([16, 0, 0, 0]);
 
@@ -154,7 +157,7 @@ impl Tag {
     /// The tag for the UTCTime type, UNIVERSAL 23.
     pub const UTC_TIME: Self = Tag([23, 0, 0, 0]);
 
-    /// The tag for the GeneralizedType type, UNIVERAL 24.
+    /// The tag for the GeneralizedType type, UNIVERSAL 24.
     pub const GENERALIZED_TIME: Self = Tag([24, 0, 0, 0]);
 
     /// The tag for the GraphicString type, UNIVERSAL 25.
@@ -169,8 +172,29 @@ impl Tag {
     /// The tag for the UniversalString type, UNIVERSAL 28.
     pub const UNIVERSAL_STRING: Self = Tag([28, 0, 0, 0]);
 
-    /// The tag for the BMPString type, UNIVERSAL 29.
-    pub const BMP_STRING: Self = Tag([29, 0, 0, 0]);
+    /// The tag for the CHARACTER STRING type, UNIVERSAL 29.
+    pub const CHARACTER_STRING: Self = Tag([29, 0, 0, 0]);
+
+    /// The tag for the BMPString type, UNIVERSAL 30.
+    pub const BMP_STRING: Self = Tag([30, 0, 0, 0]);
+
+    /// The tag for the DATE type, UNIVERSAL 31.
+    pub const DATE: Self = Tag([31, 0, 0, 0]);
+
+    /// The tag for the TIME-OF-DAY type, UNIVERSAL 32.
+    pub const TIME_OF_DAY: Self = Tag([32, 0, 0, 0]);
+
+    /// The tag for the DATE-TIME type, UNIVERSAL 33.
+    pub const DATE_TIME: Self = Tag([33, 0, 0, 0]);
+
+    /// The tag for the DURATION type, UNIVERSAL 34.
+    pub const DURATION: Self = Tag([34, 0, 0, 0]);
+
+    /// The tag for the OID-IRI type, UNIVERSAL 35.
+    pub const OID_IRI: Self = Tag([35, 0, 0, 0]);
+
+    /// The tag for the RELATIVE-OID-IRI type, UNIVERSAL 36.
+    pub const RELATIVE_OID_IRI: Self = Tag([36, 0, 0, 0]);
 
     //--- The first few context-specific tags.
     //
@@ -316,7 +340,9 @@ impl Tag {
 
     /// Returns the number of the tag.
     pub fn number(self) -> u32 {
-        if (Tag::SINGLEBYTE_DATA_MASK & self.0[0]) != Tag::SINGLEBYTE_DATA_MASK {
+        if (Tag::SINGLEBYTE_DATA_MASK & self.0[0])
+            != Tag::SINGLEBYTE_DATA_MASK
+        {
             // It's a single byte identifier
             u32::from(Tag::SINGLEBYTE_DATA_MASK & self.0[0])
         } else if Tag::LAST_OCTET_MASK & self.0[1] == 0 {
@@ -335,15 +361,17 @@ impl Tag {
         }
     }
 
-    /// Takes a tag from the beginning of a source.
+    /// Takes an optional tag from the beginning of a source.
     ///
     /// Upon success, returns both the tag and whether the value is
-    /// constructed. If there are no more octets available in the source,
-    /// an error is returned.
-    pub fn take_from<S: decode::Source>(
+    /// constructed.
+    pub fn take_opt_from<S: Source>(
         source: &mut S,
-    ) -> Result<(Self, bool), S::Err> {
-        let byte = source.take_u8()?;
+    ) -> Result<Option<(Self, bool)>, DecodeError<S::Error>> {
+        let byte = match source.take_opt_u8()? {
+            Some(byte) => byte,
+            None => return Ok(None)
+        };
         // clear constructed bit
         let mut data = [byte & !Tag::CONSTRUCTED_MASK, 0, 0, 0];
         let constructed = byte & Tag::CONSTRUCTED_MASK != 0;
@@ -351,13 +379,29 @@ impl Tag {
             for i in 1..=3 {
                 data[i] = source.take_u8()?;
                 if data[i] & Tag::LAST_OCTET_MASK == 0 {
-                    return Ok((Tag(data), constructed));
+                    return Ok(Some((Tag(data), constructed)));
                 }
             }
         } else {
-            return Ok((Tag(data), constructed));
+            return Ok(Some((Tag(data), constructed)));
+        }
+        Err(source.content_err(
+            "tag values longer than 4 bytes not implemented"
+        ))
+    }
+
+    /// Takes a tag from the beginning of a source.
+    ///
+    /// Upon success, returns both the tag and whether the value is
+    /// constructed. If there are no more octets available in the source,
+    /// an error is returned.
+    pub fn take_from<S: Source>(
+        source: &mut S,
+    ) -> Result<(Self, bool), DecodeError<S::Error>> {
+        match Self::take_opt_from(source)? {
+            Some(res) => Ok(res),
+            None => Err(source.content_err("additional values expected"))
         }
-        xerr!(Err(decode::Error::Unimplemented.into()))
     }
 
     /// Takes a tag from the beginning of a resource if it matches this tag.
@@ -365,10 +409,10 @@ impl Tag {
     /// If there is no more data available in the source or if the tag is
     /// something else, returns `Ok(None)`. If the tag matches `self`, returns
     /// whether the value is constructed.
-    pub fn take_from_if<S: decode::Source>(
+    pub fn take_from_if<S: Source>(
         self,
         source: &mut S,
-    ) -> Result<Option<bool>, S::Err> {
+    ) -> Result<Option<bool>, DecodeError<S::Error>> {
         if source.request(1)? == 0 {
             return Ok(None)
         }
@@ -378,9 +422,9 @@ impl Tag {
         if (data[0] & Tag::SINGLEBYTE_DATA_MASK) == Tag::SINGLEBYTE_DATA_MASK {
             let mut i = 1;
             loop {
-                if source.request(i + 1)? == 0 {
+                if source.request(i + 1)? <= i {
                     // Not enough data for a complete tag.
-                    xerr!(return Err(decode::Error::Malformed.into()))
+                    return Err(source.content_err("short tag value"))
                 }
                 data[i] = source.slice()[i];
                 if data[i] & Tag::LAST_OCTET_MASK == 0 {
@@ -388,15 +432,20 @@ impl Tag {
                 }
                 // We don’t support tags larger than 4 bytes.
                 if i == 3 {
-                    xerr!(return Err(decode::Error::Unimplemented.into()))
+                    return Err(source.content_err(
+                        "tag values longer than 4 bytes not implemented"
+                    ))
                 }
                 i += 1;
             }
         }
-        let (tag, compressed) = (Tag(data), byte & Tag::CONSTRUCTED_MASK != 0);
+        let (tag, constructed) = (
+            Tag(data),
+            byte & Tag::CONSTRUCTED_MASK != 0
+        );
         if tag == self {
-            source.advance(tag.encoded_len())?;
-            Ok(Some(compressed))
+            source.advance(tag.encoded_len());
+            Ok(Some(constructed))
         }
         else {
             Ok(None)
@@ -451,6 +500,7 @@ impl fmt::Display for Tag {
             Tag::EMBEDDED_PDV => write!(f, "EMBEDDED PDV"),
             Tag::UTF8_STRING => write!(f, "UTF8String"),
             Tag::RELATIVE_OID => write!(f, "RELATIVE-OID"),
+            Tag::TIME => write!(f, "TIME"),
             Tag::SEQUENCE => write!(f, "SEQUENCE"),
             Tag::SET => write!(f, "SET"),
             Tag::NUMERIC_STRING => write!(f, "NumericString"),
@@ -464,7 +514,14 @@ impl fmt::Display for Tag {
             Tag::VISIBLE_STRING => write!(f, "VisibleString"),
             Tag::GENERAL_STRING => write!(f, "GeneralString"),
             Tag::UNIVERSAL_STRING => write!(f, "UniversalString"),
+            Tag::CHARACTER_STRING => write!(f, "CHARACTER STRING"),
             Tag::BMP_STRING => write!(f, "BMPString"),
+            Tag::DATE => write!(f, "DATE"),
+            Tag::TIME_OF_DAY => write!(f, "TIME-OF-DAY"),
+            Tag::DATE_TIME => write!(f, "DATE-TIME"),
+            Tag::DURATION => write!(f, "DURATION"),
+            Tag::OID_IRI => write!(f, "OID-IRI"),
+            Tag::RELATIVE_OID_IRI => write!(f, "RELATIVE-OID-IRI"),
             tag => {
                 match tag.0[0] & Tag::CLASS_MASK {
                     Tag::UNIVERSAL => write!(f, "[UNIVERSAL ")?,
@@ -490,6 +547,7 @@ impl fmt::Debug for Tag {
 #[cfg(test)]
 mod test {
     use super::*;
+    use crate::decode::IntoSource;
 
     const TYPES: &[u8] = &[Tag::UNIVERSAL, Tag::APPLICATION, Tag::CONTEXT_SPECIFIC, Tag::PRIVATE];
 
@@ -503,10 +561,15 @@ mod test {
             for i in range.clone() {
                 let tag = Tag::new(typ, i);
                 let expected = Tag([typ | i as u8, 0, 0, 0]);
-                let decoded = Tag::take_from(&mut &tag.0[..]).unwrap();
-                assert_eq!(tag.take_from_if(&mut &tag.0[..]), Ok(Some(false)));
+                let decoded = Tag::take_from(
+                    &mut tag.0.into_source()
+                ).unwrap();
+                assert_eq!(
+                    tag.take_from_if(&mut tag.0.into_source()).unwrap(),
+                    Some(false)
+                );
                 // The value is not constructed.
-                assert_eq!(decoded.1, false);
+                assert!(!decoded.1);
                 // The tag is the same
                 assert_eq!(decoded.0, expected);
                 // We get the same number back.
@@ -532,10 +595,20 @@ mod test {
                 let expected = Tag([
                         Tag::SINGLEBYTE_DATA_MASK | typ, i as u8, 0, 0
                 ]);
-                let decoded = Tag::take_from(&mut &tag.0[..]).unwrap();
-                assert_eq!(tag.take_from_if(&mut &tag.0[..]), Ok(Some(false)));
+                let decoded = Tag::take_from(
+                    &mut tag.0.into_source()
+                ).unwrap();
+                assert_eq!(
+                    tag.take_from_if(&mut tag.0.into_source()).unwrap(),
+                    Some(false)
+                );
+                assert!(
+                    tag.take_from_if(
+                        &mut tag.0[0..1].into_source()
+                    ).is_err(),
+                );
                 // The value is not constructed.
-                assert_eq!(decoded.1, false);
+                assert!(!decoded.1);
                 // The tag is the same
                 assert_eq!(decoded.0, expected);
                 assert_eq!(tag.number(), i);
@@ -561,10 +634,20 @@ mod test {
                     i as u8 & !Tag::LAST_OCTET_MASK,
                     0
                 ]);
-                let decoded = Tag::take_from(&mut &tag.0[..]).unwrap();
-                assert_eq!(tag.take_from_if(&mut &tag.0[..]), Ok(Some(false)));
+                let decoded = Tag::take_from(
+                    &mut tag.0.into_source()
+                ).unwrap();
+                assert_eq!(
+                    tag.take_from_if(&mut tag.0.into_source()).unwrap(),
+                    Some(false)
+                );
+                assert!(
+                    tag.take_from_if(
+                        &mut tag.0[0..2].into_source()
+                    ).is_err(),
+                );
                 // The value is not constructed.
-                assert_eq!(decoded.1, false);
+                assert!(!decoded.1);
                 // The tag is the same
                 assert_eq!(decoded.0, expected);
                 assert_eq!(tag.number(), i);
@@ -590,10 +673,20 @@ mod test {
                     (i >> 7) as u8 | Tag::LAST_OCTET_MASK,
                     i as u8 & !Tag::LAST_OCTET_MASK
                 ]);
-                let decoded = Tag::take_from(&mut &tag.0[..]).unwrap();
-                assert_eq!(tag.take_from_if(&mut &tag.0[..]), Ok(Some(false)));
+                let decoded = Tag::take_from(
+                    &mut tag.0.into_source()
+                ).unwrap();
+                assert_eq!(
+                    tag.take_from_if(&mut tag.0.into_source()).unwrap(),
+                    Some(false)
+                );
+                assert!(
+                    tag.take_from_if(
+                        &mut tag.0[0..3].into_source()
+                    ).is_err(),
+                );
                 // The value is not constructed.
-                assert_eq!(decoded.1, false);
+                assert!(!decoded.1);
                 // The tag is the same
                 assert_eq!(decoded.0, expected);
                 assert_eq!(tag.number(), i);
@@ -607,14 +700,12 @@ mod test {
         let large_tag = [
             0b1111_1111, 0b1000_0000, 0b1000_0000, 0b1000_0000, 0b1000_0000
         ];
-        assert_eq!(
-            Tag::take_from(&mut &large_tag[..]),
-            Err(decode::Error::Unimplemented)
+        assert!(
+            Tag::take_from(&mut large_tag.into_source()).is_err()
         );
         let short_tag = [0b1111_1111, 0b1000_0000];
-        assert_eq!(
-            Tag::take_from(&mut &short_tag[..]),
-            Err(decode::Error::Malformed)
+        assert!(
+            Tag::take_from(&mut short_tag.into_source()).is_err()
         );
     }
 }
